/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Pierre Garcia
 * Contributor(s): 
 */

package org.objectweb.dream.cache.lib;

import org.objectweb.dream.util.LoggerUtil;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.Attribute;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.fractal.fraclet.annotation.annotations.type.Cardinality;
import org.objectweb.fractal.fraclet.annotation.annotations.type.Contingency;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;

import org.objectweb.dream.cache.api.CacheCapacityEvent;
import org.objectweb.dream.cache.api.CacheCapacityEventListener;
import org.objectweb.dream.cache.api.CacheEntry;
import org.objectweb.dream.cache.api.CacheEntryFactory;
import org.objectweb.dream.cache.api.CacheEntryFilter;
import org.objectweb.dream.cache.api.CacheEvent;
import org.objectweb.dream.cache.api.CacheEventListener;
import org.objectweb.dream.cache.api.CacheException;
import org.objectweb.dream.cache.api.CacheManager;
import org.objectweb.dream.cache.api.FixableCacheEntry;
import org.objectweb.dream.cache.api.InvalidCacheEntryException;
import org.objectweb.dream.cache.api.UnbindManager;
import org.objectweb.dream.cache.lib.BackgroundCleaner;
import org.objectweb.dream.cache.replacement.api.ReplacementManager;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import sun.reflect.generics.reflectiveObjects.NotImplementedException;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Collection;
import java.util.ArrayList;
import java.util.Collections;
import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;
import java.lang.ref.WeakReference;

/**
 * This class is an implementation of a cache. The data structure is a Hahsed
 * map between cache entry identifier and an entry instance. This class is a
 * fractal component.
 * 
 * @author S.Chassande-Barrioz, E.Bruneton
 * @author P.Garcia
 */
@DreamComponent
public class BasicCacheManager
    implements
      CacheManager,
      UnbindManager,
      BasicCacheManagerAttributeController
{

  public final static String                            DEFAULT_AUTO_CLEAN_SIZE          = "7%";

  // --------------------------------------------------------------------------
  // Attribute
  // --------------------------------------------------------------------------

  /**
   * This field is the max size of the cache.
   */
  @Attribute(argument = "cacheMaxSize")
  protected int                                         cacheMaxSize;

  /**
   * Threshold of the cache in percentage.
   */
  @Attribute(argument = "autoCleanThreshold")
  protected String                                      autoCleanThresHold;

  // --------------------------------------------------------------------------
  // Client Interfaces
  // --------------------------------------------------------------------------

  public final static String                            CACHE_ENTRY_FACTORY_BINDING      = "cache-entry-factory";
  public final static String                            REPLACEMENT_MANAGER_BINDING      = "replacement-manager";
  public final static String                            CACHE_LISTENER_BINDING           = "cache-listeners";
  public final static String                            CACHE_CAPACTITY_LISTENER_BINDING = "cache-capacity-listeners";

  /**
   * This field is the factory of CacheEntry instances.
   */
  @Requires(name = CACHE_ENTRY_FACTORY_BINDING)
  protected CacheEntryFactory                           cef;

  /**
   * Replacement Manager : FIFO, LRU, MRU
   */
  @Requires(name = REPLACEMENT_MANAGER_BINDING)
  protected ReplacementManager                          rm;

  /**
   * This field is the list of CacheEventListener registered.
   */
  @Requires(name = CACHE_LISTENER_BINDING, cardinality = Cardinality.COLLECTION, contingency = Contingency.OPTIONAL)
  protected HashMap<String, CacheEventListener>         cels;

  /**
   * This field is the list of CacheEventListener registered.
   */
  @Requires(name = CACHE_CAPACTITY_LISTENER_BINDING, cardinality = Cardinality.COLLECTION, contingency = Contingency.OPTIONAL)
  protected HashMap<String, CacheCapacityEventListener> ccels;

  // --------------------------------------------------------------------------
  // Service
  // --------------------------------------------------------------------------

  /**
   * Reference to this component
   */
  @Service
  Component                                             ref;

  /**
   * Logger of the component
   */
  @DreamMonolog
  protected Logger                                      logger;

  // --------------------------------------------------------------------------
  // Fields
  // --------------------------------------------------------------------------

  /**
   * This field is the cache. This is a Map which the key are object identifiers
   * and the values are FixableCacheEntry instances.
   */
  protected Map<Object, WeakCacheEntry>                 cache;

  /**
   * Queue used to register the weak references stored in the cache.
   */
  protected ReferenceQueue                  queue;

  /**
   * Thread responsible of cache cleaning
   */
  private BackgroundCleaner                             bgcleaner;

  /**
   * Threshold of the cache : number of elements. This filed is related to the
   * {@link #autoCleanThresHold} attribute
   */
  protected int                                         cacheThresHold                   = 0;

  private String                                        autoCleanSizeStr;

  private int                                           autoCleanSize;

  private boolean                                       started                          = false;

  /**
   * It builds a BasicCacheManager with an empty CacheEventListener list, an
   * empty cache and the default max size.
   */
  public BasicCacheManager()
  {
    cels = new HashMap<String, CacheEventListener>();
    ccels = new HashMap<String, CacheCapacityEventListener>();
    cache = new HashMap<Object, WeakCacheEntry>();
    queue = new ReferenceQueue();
    cacheMaxSize = NO_LIMIT;
    started = false;
    autoCleanSizeStr = null;
    cacheThresHold = 0;
  }

  // IMPLEMENTATION OF THE LifeCycleController INTERFACE //
  // ----------------------------------------------------//

  @LifeCycle(on = LifeCycleType.START)
  public void startFc() throws IllegalLifeCycleException
  {
    try
    {
      if (!started)
      {
        started = true;

        autoCleanSize = getValue(autoCleanSizeStr);
        cacheThresHold = getValue(autoCleanThresHold);
      }
      CacheCapacityEvent ev = new CapacityEvent(
          CacheCapacityEvent.NOTIFY_CACHE_RESIZE, NO_LIMIT, this.cacheMaxSize);
      for (Iterator<CacheCapacityEventListener> it = ccels.values().iterator(); it
          .hasNext();)
      {
        (it.next()).cacheResized(ev);
      }
    }
    catch (CacheException e)
    {
      throw new IllegalLifeCycleException(
          "An error occured while starting the "
              + BasicCacheManager.class.getSimpleName() + " component :\n "
              + e.getMessage());
    }
  }

  @LifeCycle(on = LifeCycleType.STOP)
  public void stopFc()
  {
    if (started)
    {
      started = false;
      int maxObjects = this.cacheMaxSize;
      try
      {
        // resizes the cache to 0 in order to clear it.
        setCacheMaxSize(0);
      }
      catch (CacheException e)
      {
      }
      // restore the value for the future startFc()
      this.cacheMaxSize = maxObjects;
    }
  }

  // IMPLEMENTATION OF THE UserBindingControler INTERFACE //
  // -----------------------------------------------------//

  public void unbindFc(String s) throws NoSuchInterfaceException
  {
    if (CACHE_ENTRY_FACTORY_BINDING.equals(s))
    {
      cef = null;
    }
    else if (REPLACEMENT_MANAGER_BINDING.equals(s))
    {
      rm = null;
      if (bgcleaner != null)
      {
        bgcleaner.stop();
        bgcleaner = null;
      }
    }
    else if (s.startsWith(CACHE_LISTENER_BINDING))
    {
      cels.remove(s);
    }
    else if (s.startsWith(CACHE_CAPACTITY_LISTENER_BINDING))
    {
      ccels.remove(s);
    }
    else
    {
      throw new NoSuchInterfaceException(s);
    }
  }

  // --------------------------------------------------------------------------
  // IMPLEMENTATION OF THE BasicCacheManagerAttributeControler INTERFACE
  // --------------------------------------------------------------------------

  public void setCacheMaxSize(int size) throws IllegalArgumentException,
      CacheException
  {
    if (logger != null && logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, "resize the cache: "
          + (size == NO_LIMIT ? "UNLIMITED" : "" + size));
    if (size == NO_LIMIT)
    {
      cacheMaxSize = NO_LIMIT;
      return;
    }
    if (size < 0)
      throw new IllegalArgumentException(
          "Impossible to set the cache size to a negative value: " + size);
    synchronized (cache)
    {
      CacheCapacityEvent ev = new CapacityEvent(
          CacheCapacityEvent.NOTIFY_CACHE_RESIZE, cacheMaxSize, size);
      if (logger.isLoggable(BasicLevel.DEBUG))
      {
        logger.log(BasicLevel.DEBUG, "resize the cache: " + cacheMaxSize
            + " to " + size + " (Nd elem=" + cache.size() + ")");
      }
      if (started)
      {
        // if this component is stopping
        // you cannot access to the replacement manager
        int delta = size - cache.size();
        if (delta < 0)
        {
          logger.log(BasicLevel.DEBUG, "free the cache forced");
          delta = -delta;
          int free = rm.forceFree(delta);
          if (delta > free)
          {
            logger.log(BasicLevel.WARN, "It stays " + (delta - free)
                + " entries which have not been freed");
          }
          logger.log(BasicLevel.DEBUG, "cache freed");
        }
      }
      cacheMaxSize = size;
      if (started)
      {
        for (Iterator<CacheCapacityEventListener> it = ccels.values()
            .iterator(); it.hasNext();)
        {
          (it.next()).cacheResized(ev);
        }
      }
      else
      {
        // the composant has been stopped, rehash the data structure
        Map<Object, WeakCacheEntry> old = cache;
        cache = new HashMap<Object, WeakCacheEntry>(cacheMaxSize, 1);
        if (old.size() > 0)
        {
          cache.putAll(old);
        }
      }
    }
    if (logger != null && logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, "cache resized to : "
          + (size == NO_LIMIT ? "UNLIMITED" : "" + size));
  }

  public int getCacheMaxSize()
  {
    return cacheMaxSize;
  }

  public int getCurrentSize()
  {
    return cache.size();
  }

  public Collection<Object> getCurrentEntryIdentifiers()
  {
    synchronized (cache)
    {
      if (cache.size() == 0)
      {
        return Collections.emptySet();
      }
      return Collections.unmodifiableCollection(cache.keySet());
    }
  }

  public void setMemorySize(int size) throws IllegalArgumentException
  {
    if (size < 0)
    {
      throw new IllegalArgumentException(
          "Impossible to set the cache size to a negative value: " + size);
    }
    throw new NotImplementedException();
  }

  public int getMemorySize()
  {
    throw new NotImplementedException();
  }

  public int getCurrentMemorySize()
  {
    throw new NotImplementedException();
  }

  public void setAutoCleanSize(String size) throws CacheException
  {
    autoCleanSizeStr = size;
    autoCleanSize = getValue(autoCleanSizeStr);
  }

  public String getAutoCleanSize()
  {
    return autoCleanSizeStr;
  }

  public void setAutoCleanThreshold(String size) throws CacheException
  {
    autoCleanThresHold = size;
    cacheThresHold = getValue(autoCleanThresHold);
  }

  public String getAutoCleanThreshold()
  {
    return autoCleanThresHold;
  }

  private int getValue(String str) throws CacheException
  {
    if (cacheMaxSize == NO_LIMIT)
    {
      return 0;
    }
    if (str == null)
    {
      str = "75%";
    }
    int i = str.indexOf('%');
    if (i == -1)
    {
      return Integer.parseInt(str);
    }
    else
    {
      int res = Integer.parseInt(str.substring(0, i));
      if (res < 0 || res > 100)
        throw new CacheException(
            "The auto-clean threashold must be a percentage (e. g. 50%) or a number of element. "
                + res + " is not a correct value");
      return (cacheMaxSize * res) / 100;
    }
  }

  // IMPLEMENTATION OF THE CacheManager INTERFACE //
  // ---------------------------------------------//

  public CacheEntry bind(Object id, Object object) throws CacheException
  {
    if (id == null || object == null)
      throw new IllegalArgumentException(
          "Impossible to add a cache entry with a null identifier or object");
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, "bind an element in the cache, id:" + id);
    synchronized (cache)
    {
      clean();

      // checks that the cache is not full
      int size = cache.size();
      if (cacheMaxSize != NO_LIMIT)
      {
        if (size >= cacheMaxSize)
        {
          // removes one element
          rm.forceFree(autoCleanSize);
          if (cache.size() >= cacheMaxSize)
          {
            // if the cache is still full, calls gc and waits for some time
            forceClean();
            size = cache.size();
            if (size >= cacheMaxSize)
            {
              // if the cache is really still full, log a warning
              logger.log(BasicLevel.WARN,
                  "The cache is still full after cleaning: number of exceeded entries is "
                      + (size - cacheMaxSize));
              // throw new CacheFullException();
            }
          }
        }
        else if ((size + 1) >= cacheThresHold && getBgcleaner() != null)
        {
          getBgcleaner().backgroungCleanup(autoCleanSize);
        }
      }

      // checks that 'id' is not already bound
      WeakCacheEntry ref = (WeakCacheEntry) cache.get(id);
      if (ref != null)
      {
        FixableCacheEntry entry = (FixableCacheEntry) ref.get();
        // if (entry != null && !(entry.getCeObject() == object && ref.ce ==
        // null)) {
        if (entry != null && ref.ce != null)
        {
          throw new CacheException("Identifier is already bound with "
              + (entry.getCeObject() == object ? "the same" : "another")
              + " object: \nid=" + id);
        }
      }

      // creates a new entry in the cache
      FixableCacheEntry ce = cef.create(id, object);
      cache.put(id, new WeakCacheEntry(ce, queue));
      rm.addForReplacement(ce);
      CacheEvent ev = new Event(CacheEvent.NOTIFY_BIND, ce, cache.size());
      logger.log(BasicLevel.DEBUG, "notify the bind to cache event listeners");
      for (Iterator<CacheEventListener> it = cels.values().iterator(); it
          .hasNext();)
      {
        (it.next()).entryBound(ev);
      }
      return ce;
    }
  }

  public CacheEntry lookup(Object id)
  {
    if (id == null)
      throw new IllegalArgumentException(
          "Impossible to find cache entry with a null identifier");
    synchronized (cache)
    {
      try
      {
        CacheEntry ce = get(id);
        if (logger.isLoggable(BasicLevel.DEBUG))
        {
          logger.log(BasicLevel.DEBUG, "lookup: id=" + id);
        }
        return ce;
      }
      catch (InvalidCacheEntryException e)
      {
        return null;
      }
    }
  }

  public void fix(FixableCacheEntry _ce) throws CacheException
  {
    if (_ce == null)
      throw new IllegalArgumentException("Impossible to fix a null cache entry");
    synchronized (_ce)
    {
      FixableCacheEntry ce = (FixableCacheEntry) get(_ce.getCeIdentifier());
      ce.fixCe();
      if (logger.isLoggable(BasicLevel.DEBUG))
        logger.log(BasicLevel.DEBUG, "Fix an element in the cache, id:"
            + ce.getCeIdentifier() + " / count:" + ce.getCeFixCount());
    }
  }

  public void fix(CacheEntryFilter filter) throws CacheException
  {
    synchronized (cache)
    { // to prevent lookup, bind in concurrence
      for (Iterator<WeakCacheEntry> iter = cache.values().iterator(); iter
          .hasNext();)
      {
        CacheEntry ce = iter.next().ce;
        if (filter.accept(ce))
        {
          fix((FixableCacheEntry) ce);
        }
      }
    }
  }

  public void unfix(FixableCacheEntry ce) throws CacheException
  {
    if (ce == null)
      throw new IllegalArgumentException(
          "Impossible to unfix a null cache entry");
    synchronized (ce)
    {
      ((FixableCacheEntry) ce).unfixCe();
      if (logger.isLoggable(BasicLevel.DEBUG))
        logger.log(BasicLevel.DEBUG, "UnFix an element in the cache, id:"
            + ce.getCeIdentifier() + " / count:"
            + ((FixableCacheEntry) ce).getCeFixCount());
    }
  }

  public void unfix(CacheEntryFilter filter) throws CacheException
  {
    synchronized (cache)
    { // to prevent lookup, bind in concurrence
      for (Iterator<WeakCacheEntry> iter = cache.values().iterator(); iter
          .hasNext();)
      {
        CacheEntry ce = iter.next().ce;
        if (filter.accept(ce))
        {
          unfix((FixableCacheEntry) ce);
        }
      }
    }
  }

  public void touch(CacheEntry ce) throws CacheException
  {
    if (ce == null)
      throw new IllegalArgumentException(
          "Impossible to touch a null cache entry");
    synchronized (ce)
    {
      // Check that the entry exists in the cache
      rm.adjustForReplacement((FixableCacheEntry) ce);
    }
  }

  public void unbindAll() throws CacheException
  {
    synchronized (cache)
    {
      rm.unbindAll(true);
    }
  }

  // IMPLEMENTATION OF THE UnbindManager INTERFACE //
  // ----------------------------------------------//

  public boolean unbind(Object oid, boolean force) throws CacheException
  {
    synchronized (cache)
    { // to prevent lookup, bind in concurrence
      return _unbind(oid, force);
    }
  }

  private boolean _unbind(Object oid, boolean force) throws CacheException
  {
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, "Unbind an element from the cache, id:"
          + oid);
    clean();
    WeakCacheEntry ref = (WeakCacheEntry) cache.get(oid);
    if (ref == null)
    {
      return true;
    }
    FixableCacheEntry ce = (FixableCacheEntry) ref.get();
    if (ce == null)
    {
      throw new InvalidCacheEntryException();
    }
    synchronized (ce)
    {
      if (ce.getCeFixCount() > 0)
      {
        return false;
      }
      // removes the strong reference to the real entry,
      // but does not remove it from 'cache'
      ref.ce = null;
      if (force)
      {
        cache.remove(oid);
        CacheEvent ev = new Event(CacheEvent.NOTIFY_UNBIND, oid, cache.size());
        for (Iterator<CacheEventListener> it = cels.values().iterator(); it
            .hasNext();)
        {
          (it.next()).entryUnbound(ev);
        }
        rm.removeForReplacement(oid);
      }
      return force;
    }
  }

  public Collection<Object> unbind(CacheEntryFilter filter, boolean force)
      throws CacheException
  {
    Collection<Object> res = new ArrayList<Object>();
    synchronized (cache)
    { // to prevent lookup, bind in concurrence
      for (Iterator<WeakCacheEntry> iter = cache.values().iterator(); iter
          .hasNext();)
      {
        CacheEntry ce = (CacheEntry) iter.next().ce;
        if (filter.accept(ce))
        {
          Object oid = ce.getCeIdentifier();
          if (_unbind(ce, force))
          {
            res.add(oid);
          }
        }
      }
    }
    return res;
  }

  public Collection<Object> unbindAll(Collection<Object> oids, boolean force)
      throws CacheException
  {
    ArrayList<Object> unbound = new ArrayList<Object>(oids.size());
    synchronized (cache)
    { // to prevent lookup, bind in concurrence
      if (cache.size() == 0)
      {
        unbound.addAll(oids);
        return unbound;
      }
      Iterator<Object> it = oids.iterator();
      while (it.hasNext())
      {
        Object oid = it.next();
        if (!unbound.contains(oid)
            && (lookup(oid) == null || _unbind(oid, force)))
        {
          unbound.add(oid);
        }
      }
      if (!force)
      {
        /*
         * In case of force == false, we run the GC in order to really evict the
         * entries
         */
        forceClean();
        clean();
        if (cache.size() == 0)
        {
          // Optimization when the user has asked an evict all on the
          // cache
          unbound = new ArrayList<Object>(oids);
          return unbound;
        }
        // Check if some entries has been removed
        it = oids.iterator();
        while (it.hasNext())
        {
          Object oid = it.next();
          WeakCacheEntry ref = (WeakCacheEntry) cache.get(oid);
          if (ref == null)
          {
            if (!unbound.contains(oid))
            {
              unbound.add(oid);
            }
          }
        }
      }
    }
    return unbound;
  }

  public void unbindAll(boolean force) throws CacheException
  {
    synchronized (cache)
    { // to prevent lookup, bind in concurrence
      // clear the cache
      cache.clear();
      if (!force)
      {
        /*
         * In case of force == false, we run the GC in order to really evict the
         * entries
         */
        forceClean();
        clean();
      }
    }
    LoggerUtil.end(logger);
  }

  public Collection<Object> unbindUnfixed(boolean force) throws CacheException
  {
    throw new NotImplementedException();
    // return Collections.EMPTY_SET;
  }

  // ----------------------------------------------//

  private CacheEntry get(Object id) throws InvalidCacheEntryException
  {
    WeakCacheEntry ref = cache.get(id);
    if (ref == null)
    {
      /*
       * The cache can be resized during the call, and the entry can disappear
       * temporaly. Then to really known if the entry does not exist a
       * synchrnonization of the cache is neeeded.
       */
      synchronized (cache)
      {
        ref = cache.get(id);
      }
      if (ref == null)
      {
        throw new InvalidCacheEntryException(id);
      }
    }
    CacheEntry entry = ref.get();
    if (entry == null)
    {
      throw new InvalidCacheEntryException();
    }
    // Re-born the entry ;-)
    ref.ce = entry;
    return entry;
  }

  /**
   * Removes all entries that have been garbage collected. Does not block.
   */
  private static final int NB_CALL_BEFORE_ACTUAL_CLEAN = 2;
  private int              nbClean                     = 0;

  private void clean() throws CacheException
  {
    if (cacheMaxSize != NO_LIMIT)
    {
      if (cache.size() < cacheThresHold)
      {
        if (nbClean < NB_CALL_BEFORE_ACTUAL_CLEAN)
        {
          nbClean++;
          return;
        }
      }
    }
    else
    {
      if (nbClean < NB_CALL_BEFORE_ACTUAL_CLEAN)
      {
        nbClean++;
        return;
      }
    }
    nbClean = 0;
    while (true)
    {
      Reference<? extends CacheEntry> ref = queue.poll();
      if (ref != null)
      {
        CacheEntry ce = ref.get();
        if (ce != null)
        {
          if (logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "Really remove from the cache, id:"
                + ce.getCeIdentifier());
          Object oid = ce.getCeIdentifier();
          cache.remove(oid);
          CacheEvent ev = new Event(CacheEvent.NOTIFY_UNBIND, oid, cache.size());
          for (Iterator<CacheEventListener> it = cels.values().iterator(); it
              .hasNext();)
          {
            (it.next()).entryUnbound(ev);
          }
          rm.removeForReplacement(oid);
        }
      }
      else
      {
        return;
      }
    }
  }

  /**
   * Calls the garbage collector and waits until an entry can be removed, or a
   * timeout period (TIMEOUTINMILLISECONDS milli-seconds) has expired.
   */
  private static final int TIMEOUTINMILLISECONDS = 50;

  private void forceClean()
  {
    logger.log(BasicLevel.DEBUG, "forceClean()");
    for (int i = 0; i < 6; ++i)
    {
      System.gc();
      WeakCacheEntry ref = null;
      try
      {
        ref = (WeakCacheEntry) queue.remove(TIMEOUTINMILLISECONDS);
      }
      catch (InterruptedException e)
      {
      }
      if (ref != null && ref.ce == null)
      {
        Object oid = ref.getIdentifier();
        cache.remove(oid);
        CacheEvent ev = new Event(CacheEvent.NOTIFY_UNBIND, oid, cache.size());
        for (Iterator<CacheEventListener> it = cels.values().iterator(); it
            .hasNext();)
        {
          (it.next()).entryUnbound(ev);
        }
        rm.removeForReplacement(oid);
        return;
      }
    }
  }

  /**
   * The entries stored in 'cache' are weak references to the real entries. Each
   * weak entry also stores the id of the entry (this id is used in clean and
   * forceClean to remove the entry from 'cache')
   */
  private static class WeakCacheEntry extends WeakReference<CacheEntry>
  {

    private final Object id;

    // strong reference to the real entry: prevents it from being garbage
    // collected until the unbind method is called.
    public CacheEntry    ce;

    public WeakCacheEntry(CacheEntry ce, ReferenceQueue<CacheEntry> queue)
    {
      super(ce, queue);
      this.id = ce.getCeIdentifier();
      this.ce = ce;
    }

    public Object getIdentifier()
    {
      return id;
    }
  }

  /**
   * This class is a basic implementation of the CacheEvent interface.
   */
  protected class Event implements CacheEvent
  {

    public int        eventId   = 0;
    public CacheEntry ce        = null;
    public Object     oid       = null;
    public int        cacheSize = -2;

    public Event(int eventid, Object oid, int cachesize)
    {
      eventId = eventid;
      this.oid = oid;
      cacheSize = cachesize;
    }

    public Event(int eventid, CacheEntry _ce, int cachesize)
    {
      eventId = eventid;
      ce = _ce;
      cacheSize = cachesize;
    }

    public int getEventId()
    {
      return eventId;
    }

    public Object getCeIdentifier()
    {
      return oid;
    }

    public CacheEntry getEntry()
    {
      return ce;
    }
  }

  /**
   * This class is a basic implementation of the CacheCapacityEvent interface.
   */
  protected class CapacityEvent implements CacheCapacityEvent
  {
    public int eventId      = 0;
    public int oldCacheSize = -2;
    public int cacheSize    = -2;

    public CapacityEvent(int eventid, int oldCacheSize, int cacheSize)
    {
      this.eventId = eventid;
      this.cacheSize = cacheSize;
      this.oldCacheSize = oldCacheSize;
    }

    public int getEventId()
    {
      return eventId;
    }

    public int getOldSize()
    {
      return oldCacheSize;
    }

    public int getSize()
    {
      return cacheSize;
    }
  }

  private BackgroundCleaner getBgcleaner()
  {
    if (bgcleaner == null && cacheThresHold > 0)
    {
      bgcleaner = new BackgroundCleaner(rm, logger);
    }
    return bgcleaner;
  }
}
