/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.control.activity.manager;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.dream.control.activity.task.IllegalTaskException;
import org.objectweb.dream.control.activity.task.Task;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;

/**
 * Routes {@link SchedulerManager }calls to the correct scheduler interface
 * depending on hints.
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
public class MetaSchedulerManager implements SchedulerManager
{

  static final String         THREAD_HINTS_KEY            = "thread";
  static final String         MONO_THREAD_HINTS_VALUE     = "mono";
  static final String         POOL_THREAD_HINTS_VALUE     = "pool";
  static final String         PERIODIC_HINTS_KEY          = "period";
  Map<Task, SchedulerManager> taskScheduler               = new HashMap<Task, SchedulerManager>();

  // ---------------------------------------------------------------------------
  // client interfaces
  // ---------------------------------------------------------------------------

  static final String         TPT_SCHEDULER_ITF_NAME      = "tpt-scheduler-manager";
  @Requires(name = TPT_SCHEDULER_ITF_NAME)
  SchedulerManager            tptSchedulerItf;

  static final String         TPPT_SCHEDULER_ITF_NAME     = "tppt-scheduler-manager";
  @Requires(name = TPPT_SCHEDULER_ITF_NAME)
  SchedulerManager            tpptSchedulerItf;

  static final String         PERIODIC_SCHEDULER_ITF_NAME = "periodic-scheduler-manager";
  @Requires(name = PERIODIC_SCHEDULER_ITF_NAME)
  SchedulerManager            periodicSchedulerItf;

  // ---------------------------------------------------------------------------
  // Implementation of the SchedulerManager interface
  // ---------------------------------------------------------------------------

  /**
   * @see SchedulerManager#scheduleTask(Task, Map)
   */
  public synchronized Object scheduleTask(Task task, Map hints) throws IllegalTaskException
  {
    SchedulerManager itf;
    Object h = hints.get(PERIODIC_HINTS_KEY);
    if (h != null)
    {
      itf = periodicSchedulerItf;
    }
    else
    {
      String s = (String) hints.get(THREAD_HINTS_KEY);
      if (s == null || MONO_THREAD_HINTS_VALUE.equals(s))
      {
        itf = tptSchedulerItf;
      }
      else if (POOL_THREAD_HINTS_VALUE.equals(s))
      {
        itf = tpptSchedulerItf;
      }
      else
      {
        throw new IllegalTaskException(task, "Invalid hints");
      }
    }
    synchronized (this)
    {
      taskScheduler.put(task, itf);
      return itf.scheduleTask(task, hints);
    }
  }

  /**
   * @see SchedulerManager#unscheduleTask(Task)
   */
  public synchronized void unscheduleTask(Task task)
      throws IllegalTaskException
  {
    SchedulerManager itf = taskScheduler.remove(task);
    if (itf == null)
    {
      throw new IllegalTaskException(task, "Unknown task");
    }
    itf.unscheduleTask(task);
  }

  /**
   * @see org.objectweb.dream.control.activity.scheduler.InterruptTask#interruptTask(Task)
   */
  public synchronized void interruptTask(Task task) throws IllegalTaskException
  {
    SchedulerManager itf = taskScheduler.get(task);
    if (itf == null)
    {
      throw new IllegalTaskException(task, "Unknown task");
    }
    itf.interruptTask(task);
  }
}
