/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.control.activity.manager;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.dream.control.activity.scheduler.ForwarderSchedulerImpl;
import org.objectweb.dream.control.activity.scheduler.InterruptTask;
import org.objectweb.dream.control.activity.scheduler.Scheduler;
import org.objectweb.dream.control.activity.task.IllegalTaskException;
import org.objectweb.dream.control.activity.task.Task;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.Logger;

/**
 * Scheduler manager that schedules each task with a
 * {@link ForwarderSchedulerImpl }and a thread.
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
public class ThreadPerTaskSchedulerManager extends AbstractSchedulerManager
    implements
      EndOfThread
{
  // --------------------------------------------------------------------------
  // Constants
  // --------------------------------------------------------------------------

  static final String  THREAD_PER_TASK_END_OF_THREAD_ITF_NAME = "tpt-end-of-thread";

  // --------------------------------------------------------------------------
  // Fields
  // --------------------------------------------------------------------------

  Map<Task, TaskState> scheduledTask                          = new HashMap<Task, TaskState>();
  Map<Task, TaskState> schedulerThread                        = new HashMap<Task, TaskState>();

  // --------------------------------------------------------------------------
  // Services interfaces
  // --------------------------------------------------------------------------

  /**
   * Component reference
   */
  @Service
  Component            weaveableC;

  /**
   * Logger used in concret implementation
   */
  @DreamMonolog()
  protected Logger     logger;

  // ---------------------------------------------------------------------------
  // Implementation of the SchedulerManager interface
  // ---------------------------------------------------------------------------

  /**
   * @see SchedulerManager#scheduleTask(Task, Map)
   */
  public synchronized Object scheduleTask(Task task, Map hints)
      throws IllegalTaskException
  {
    if (scheduledTask.containsKey(task))
    {
      throw new IllegalTaskException(task, "Task already scheduled");
    }

    // add and bind scheduler
    Scheduler scheduler = new ForwarderSchedulerImpl();
    Task thread = threadManagerItf.addThread();
    try
    {
      contentControllerItf.addFcSubComponent(scheduler);
      ((BindingController) scheduler).bindFc(Task.ITF_NAME, task);
      ((BindingController) scheduler).bindFc(EndOfTask.ITF_NAME,
          contentControllerItf.getFcInternalInterface(EndOfTask.ITF_NAME));
      ((BindingController) scheduler).bindFc("logger", logger);
      ((BindingController) thread).bindFc(Scheduler.ITF_NAME, scheduler);
      ((BindingController) thread).bindFc(EndOfThread.ITF_NAME,
          contentControllerItf
              .getFcInternalInterface(THREAD_PER_TASK_END_OF_THREAD_ITF_NAME));
      ((LifeCycleController) thread).startFc();
    }
    catch (Exception e)
    {
      Error.bug(logger, e);
    }

    TaskState taskState = new TaskState(task, scheduler, thread);
    scheduledTask.put(task, taskState);
    schedulerThread.put(thread, taskState);
    return null;
  }

  /**
   * <i>prerequisite</i> : the method {@link #threadEnded(Task)} <b>must</b>
   * have been called before.
   * 
   * @see SchedulerManager#unscheduleTask(Task)
   */
  public synchronized void unscheduleTask(Task task)
      throws IllegalTaskException
  {
    TaskState state = scheduledTask.get(task);
    if (state == null)
    {
      throw new IllegalTaskException(task, "Unknown task");
    }
    try
    {
      // EndofTask notified to task manager, just unbind scheduler
      ((BindingController) state.scheduler).unbindFc(Task.ITF_NAME);
    }
    catch (Exception e)
    {
      Error.bug(logger, e);
    }
  }

  /**
   * @see InterruptTask#interruptTask(Task)
   */
  public synchronized void interruptTask(Task task) throws IllegalTaskException
  {
    TaskState state = scheduledTask.get(task);
    if (state == null)
    {
      throw new IllegalTaskException(task, "Unknown task");
    }
    ((InterruptTask) state.scheduler).interruptTask(task);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the EndOfThread interface
  // ---------------------------------------------------------------------------

  /**
   * @see EndOfThread#threadEnded(Task)
   * @return <code>false</code> : the thread <b>must</b> never be restarted for this
   *         kind of scheduler.
   */
  public synchronized boolean threadEnded(Task thread)
  {
    TaskState state = schedulerThread.get(thread);
    if (state == null)
    {
      Error.bug(logger);
    }
    try
    {
      // don't unbind components, just remove them.
      contentControllerItf.removeFcSubComponent(state.scheduler);
      threadManagerItf.removeThread(thread);
    }
    catch (Exception e)
    {
      Error.bug(logger, e);
    }
    scheduledTask.remove(state.task);
    schedulerThread.remove(thread);
    return false;
  }

  // ---------------------------------------------------------------------------
  // inner class
  // ---------------------------------------------------------------------------

  /**
   * State of a {@link Task} containing following informations
   * <ul>
   * <li><code>scheduler</code> the scheduler associated to the task ({@link ForwarderSchedulerImpl}
   * here)</li>
   * <li><code>thread</code> the thread associated the <i>high-level</i>
   * task</li>
   * </ul>
   */
  private static class TaskState
  {
    Task      task;
    Scheduler scheduler;
    Task      thread;

    /**
     * Constructor
     * 
     * @param task related task
     * @param scheduler sheduler linked to the task
     * @param thread thread linked to the task
     */
    TaskState(Task task, Scheduler scheduler, Task thread)
    {
      this.task = task;
      this.scheduler = scheduler;
      this.thread = thread;
    }
  }
}
