/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.control.logger;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.julia.loader.Initializable;
import org.objectweb.fractal.julia.loader.Tree;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Basic implementation of logger controller. A <code>"monolog-conf-file"</code>
 * initialization parameter can be set to specify the location of the monolog
 * configuration file.
 */
public abstract class BasicLoggerControllerMixin
    implements
      LoggerController,
      LoggerControllerRegister,
      Initializable
{

  String                     baseName;

  /** Map associating logger name and set of registered loggables */
  Map<String, Set<Loggable>> registrations;
  /** Map associating logger name and logger instance */
  Map<String, Logger>        loggers;
  Logger                     baseLogger;

  // -------------------------------------------------------------------------
  // Fields and methods added and overridden by the mixin class
  // -------------------------------------------------------------------------

  /**
   * @see Initializable#initialize(org.objectweb.fractal.julia.loader.Tree)
   */
  public void initialize(final Tree args) throws Exception
  {
    _super_initialize(args);
    for (int i = 0; i < args.getSize(); ++i)
    {
      Tree arg = args.getSubTree(i);
      if (arg.getSize() == 2 && arg.getSubTree(0).equals("monolog-conf-file"))
      {
        String monologConfFile = arg.getSubTree(1).toString();
        if (Monolog.monologFactory == Monolog.getDefaultMonologFactory())
        {
          Monolog.getMonologFactory(monologConfFile);
        }
      }
    }
  }

  /**
   * @see org.objectweb.fractal.julia.Controller#initFcController(InitializationContext)
   */
  public void initFcController(InitializationContext ic)
      throws org.objectweb.fractal.api.factory.InstantiationException
  {
    _super_initFcController(ic);
    baseName = Util.getNextUnnamedBaseName();
    registrations = new HashMap<String, Set<Loggable>>();
    loggers = new HashMap<String, Logger>();
  }

  /**
   * @see LoggerController#getBaseName()
   */
  public String getBaseName()
  {
    return baseName;
  }

  /**
   * @see LoggerController#setBaseName(String)
   */
  public void setBaseName(String name)
  {
    this.baseName = name;
    baseNameChanged();
  }

  /**
   * @see LoggerController#getLoggerLevel(String)
   */
  public int getLoggerLevel(String loggerName)
  {
    Logger l = loggers.get(loggerName);
    if (l == null)
    {
      if (loggerName == null)
      {
        if (baseLogger == null)
        {
          baseLogger = Monolog.monologFactory.getLogger(baseName);
        }
        l = baseLogger;
      }
      else
      {
        return INHERIT;
      }
    }
    int i = l.getCurrentIntLevel();
    if (i == BasicLevel.DEBUG) return DEBUG;
    if (i == BasicLevel.INFO) return INFO;
    if (i == BasicLevel.WARN) return WARN;
    if (i == BasicLevel.ERROR) return ERROR;
    if (i == BasicLevel.FATAL) return FATAL;
    if (i == BasicLevel.INHERIT) return INHERIT;
    throw new IllegalStateException("Unknown logger level : " + i);
  }

  /**
   * @see LoggerController#setLoggerLevel(String, int)
   */
  public void setLoggerLevel(String loggerName, int level)
  {
    Logger l = loggers.get(loggerName);
    if (l == null)
    {
      if (loggerName == null)
      {
        if (baseLogger == null)
        {
          baseLogger = Monolog.monologFactory.getLogger(baseName);
        }
        l = baseLogger;
      }
      else
      {
        return;
      }
    }
    switch (level)
    {
      case DEBUG :
        l.setLevel(BasicLevel.LEVEL_DEBUG);
        break;
      case INFO :
        l.setLevel(BasicLevel.LEVEL_INFO);
        break;
      case WARN :
        l.setLevel(BasicLevel.LEVEL_WARN);
        break;
      case ERROR :
        l.setLevel(BasicLevel.LEVEL_ERROR);
        break;
      case FATAL :
        l.setLevel(BasicLevel.LEVEL_FATAL);
        break;
      case INHERIT :
        l.setLevel(BasicLevel.LEVEL_INHERIT);
        break;
      default :
        throw new IllegalArgumentException("Unknown level " + level);
    }
  }

  /**
   * @see LoggerController#getLoggerNames()
   */
  public String[] getLoggerNames()
  {
    return loggers.keySet().toArray(new String[loggers.size()]);
  }

  /**
   * @see LoggerControllerRegister#register(String, Loggable)
   */
  public void register(String loggerName, Loggable loggable)
  {
    Set<Loggable> s = registrations.get(loggerName);
    if (s == null)
    {
      s = new HashSet<Loggable>();
      registrations.put(loggerName, s);
    }
    s.add(loggable);
    giveLogger(loggerName, loggable);
  }

  /**
   * @see LoggerControllerRegister#unregiser(String, Loggable)
   */
  public void unregiser(String loggerName, Loggable loggable)
  {
    Set<Loggable> s = registrations.get(loggerName);
    if (s == null)
    {
      return;
    }
    s.remove(loggable);
    if (s.isEmpty())
    {
      // no more registration for this logger, remove reference to it, for
      // garbage collector
      registrations.remove(loggerName);
      loggers.remove(loggerName);
    }
  }

  //---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------
  
  /**
   * Method called when the base of a component changes 
   *
   */
  void baseNameChanged()
  {
    loggers.clear();
    Iterator<Map.Entry<String, Set<Loggable>>> iter = registrations.entrySet()
        .iterator();
    while (iter.hasNext())
    {
      Map.Entry<String, Set<Loggable>> entry = iter.next();
      String loggerName = entry.getKey();
      Set<Loggable> loggables = entry.getValue();
      Iterator<Loggable> iter2 = loggables.iterator();
      while (iter2.hasNext())
      {
        Loggable loggable = iter2.next();
        giveLogger(loggerName, loggable);
      }
    }
  }

  /**
   * Give a logger to a component (which implement {@link Loggable} interface
   * @param loggerName name of the logger (example from AbstractComponent.java : "life-cycle")
   * @param loggable a component which implements {@link Loggable} interface (example : AbstractComponent) 
   */
  void giveLogger(String loggerName, Loggable loggable)
  {
    Logger logger = loggers.get(loggerName);
    if (logger == null)
    {
      String name;
      if (loggerName == null)
      {
        name = baseName;
      }
      else
      {
        name = baseName + "." + loggerName;
      }
      logger = Monolog.monologFactory.getLogger(name);
      loggers.put(loggerName, logger);
    }
    loggable.setLogger(loggerName, logger);
  }

  // ---------------------------------------------------------------------------
  // Fields and methods required by the mixin class in the base class
  // ---------------------------------------------------------------------------

  /**
   * The
   * {@link org.objectweb.fractal.julia.Controller#initFcController(InitializationContext) initFcController}
   * method overridden by this mixin.
   * 
   * @see org.objectweb.fractal.julia.Controller#initFcController(InitializationContext)
   */
  public abstract void _super_initFcController(InitializationContext ic)
      throws org.objectweb.fractal.api.factory.InstantiationException;

  /**
   * The {@link Initializable#initialize(Tree) initialize}method overridden by
   * this mixin.
   * 
   * @see Initializable#initialize(Tree)
   */
  public abstract void _super_initialize(final Tree args) throws Exception;

}
