/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.message;

import java.io.Externalizable;
import java.lang.ref.Reference;

import org.objectweb.dream.pool.Recyclable;

/**
 * Base class of every chunk classes. This class is parameterized with a type of
 * the form <code>T extends AbstractChunk&lt;T&gt;</code>. This means that a
 * concrete chunk class must be declare as following:
 * 
 * <pre>
 * public class <i>ChunkClass</i> extends AbstractChunk&lt;<i>ChunkClass</i>&gt; { ... }
 * </pre>
 * 
 * This way, the class <code>ChunkClass</code> respects the pattern
 * <code>T extends AbstractChunk&lt;T&gt;</code>, and instances of
 * <code>ChunkClass</code> have a field of type
 * <code>ChunkFactoryReference&lt;ChunkClass&gt;</code>.
 * <p>
 * This generic type implies a limitation: any sub class of
 * <code>ChunkClass</code> is not a valid chunk type. For instance a class
 * defined as following :
 * 
 * <pre>
 * public class <i>SubChunkClass</i> extends <i>ChunkClass</i> { ... }
 * </pre>
 * 
 * does not match the pattern <code>T extends AbstractChunk&lt;T&gt;</code>,
 * since it matches
 * <code><i>SubChunkClass</i> extends AbstractChunk&lt;<i>ChunkClass</i>&gt;</code>.
 * 
 * @param <T> the concrete chunk type.
 */
public abstract class AbstractChunk<T extends AbstractChunk<T>>
    implements
      Externalizable,
      Recyclable
{
  private ChunkFactoryReference<T>         factory;
  private MessageManagerType                   owner;
  private Reference< ? extends Recyclable> recyclableReference;

  /**
   * Initialize the chunk by setting the factory and the message manager 
   * linked to this chunk 
   * @param factory reference of the associated chunk factory
   * @param owner {@link MessageManager} owner of the chunk
   */
  final void initAbstractChunk(final ChunkFactoryReference<T> factory,
      final MessageManagerType owner)
  {
    this.factory = factory;
    this.owner = owner;
  }

  /**
   * Get the factory which created this chunk. This method is useful to
   * clone a chunk
   * @return the {@link ChunkFactoryReference} linked to the chunk
   */
  final ChunkFactoryReference<T> getFactory()
  {
    return factory;
  }

  /**
   * Get the {@MessageManager} owner of the chunk. This method is useful
   * to check the owner of a chunk
   * @return a reference to the linked {@link MessageManager}
   */
  final MessageManagerType getOwner()
  {
    return owner;
  }

  /**
   * Recycle the chunk using the related chunk factory.
   */
  @SuppressWarnings("unchecked")
  final void recycleChunk()
  {
    T t = (T) this;
    factory.recycleChunk(t);
  }
  
  /**
   * Create a new chunk of type T 
   * @return the created chunk
   */
  protected abstract T newChunk();

  /**
   * Transfer the state of a chunk to another chunk
   * @param newInstance the newinstance of a chunk of type T
   */
  protected abstract void transfertStateTo(T newInstance);

  /**
   * @see Recyclable#setReference(Reference)
   */
  public void setReference(final Reference< ? extends Recyclable> reference)
  {
    this.recyclableReference = reference;
  }

  /**
   * @see Recyclable#getReference()
   */
  public Reference< ? extends Recyclable> getReference()
  {
    return recyclableReference;
  }

}
