/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.message;

import java.util.Iterator;

import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.util.EmptyIterator;
import org.objectweb.dream.util.IteratorNoRemove;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.Attribute;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.Logger;

/**
 * Basic message manager implementation.
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
public class MessageManager implements MessageManagerType
{
  private static final Iterator<Message> EMPTY_SUB_MESSAGE_ITERATOR = new EmptyIterator<Message>();
  final ChunkFactoryRepository           chunkFactoryRepository     = new ChunkFactoryRepository();

  /**
   * Component reference
   */
  @Service
  Component weaveableC;
  
  /**
   * Logger used in concret implementation
   */ 
  @DreamMonolog()
  protected Logger                   logger;

  /**
   * <code>doCheck</code> attribute value. If <code>true</code>, additional
   * (expensive) checks will be performed by the message manager. This can be
   * useful to detect bugs in a dream architecture. Default value is
   * <code>false</code>.
   */
  @Attribute(value = "false")
  public boolean                         check                      = false;

  // ---------------------------------------------------------------------------
  // Implementation of the MessageManagerType interface
  // ---------------------------------------------------------------------------

  // Message creation ----------------------------------------------------------

  /**
   * @see MessageManagerType#createMessage()
   */
  public Message createMessage()
  {
    final BasicMessageImpl message = new BasicMessageImpl(this);
    message.incrementReferenceCounter();
    return message;
  }

  /**
   * @see MessageManagerType#deleteMessage(Message)
   */
  public void deleteMessage(final Message message)
  {
    checkOwner(message);
    if (message.decrementReferenceCounter())
    {
      final Iterator<String> names = message.getNameIterator();
      while (names.hasNext())
      {
        final String chunkName = names.next();
        final AbstractChunk chunk = getChunk(message, chunkName);
        names.remove();
        deleteChunk(chunk);
      }
      final Iterator<Message> subMessages = message.getSubMessageIterator();
      while (subMessages.hasNext())
      {
        deleteMessage(subMessages.next());
      }
      recycleMessageInstance(message);
    }
  }

  /**
   * @see MessageManagerType#duplicateMessage(Message, boolean)
   */
  @SuppressWarnings("unchecked")
  public Message duplicateMessage(final Message message, final boolean clone)
  {
    checkOwner(message);
    if (clone)
    {
      final Message newMessage = createMessage();
      final Iterator<String> names = message.getNameIterator();
      while (names.hasNext())
      {
        final String chunkName = names.next();
        final AbstractChunk chunk = getChunk(message, chunkName);
        final AbstractChunk newChunk = chunk.getFactory().createChunk();
        chunk.transfertStateTo(newChunk);
        newMessage.addChunk(chunkName, newChunk);
      }
      final Iterator<Message> subMessages = message.getSubMessageIterator();
      while (subMessages.hasNext())
      {
        newMessage.addSubMessage(duplicateMessage(subMessages.next(), true));
      }
      return newMessage;
    }
    message.incrementReferenceCounter();
    return message;
  }

  // Chunk creation ------------------------------------------------------------

  /**
   * @see MessageManagerType#getChunkFactory(Class)
   */
  public synchronized <T extends AbstractChunk<T>> ChunkFactoryReference<T> getChunkFactory(
      final Class<T> chunkType)
  {
    ChunkFactoryReference<T> factory = chunkFactoryRepository
        .getFactory(chunkType);
    if (factory == null)
    {
      factory = createChunkFactory(chunkType);
      chunkFactoryRepository.addFactory(chunkType, factory);
    }
    return factory;
  }

  /**
   * @see MessageManagerType#createChunk(ChunkFactoryReference)
   */
  public <T extends AbstractChunk<T>> T createChunk(
      final ChunkFactoryReference<T> chunkFactory)
  {
    return chunkFactory.createChunk();
  }

  /**
   * @see MessageManagerType#cloneChunk(AbstractChunk)
   */
  public <T extends AbstractChunk<T>> T cloneChunk(final T chunk)
  {
    checkOwner(chunk);
    final T clone = chunk.getFactory().createChunk();
    chunk.transfertStateTo(clone);
    return clone;
  }

  /**
   * @see MessageManagerType#deleteChunk(AbstractChunk)
   */
  public void deleteChunk(final AbstractChunk<?> chunk)
  {
    checkOwner(chunk);
    chunk.recycleChunk();
  }

  // Chunk manipulation --------------------------------------------------------

  /**
   * @see MessageManagerType#isEmpty(Message)
   */
  public boolean isEmpty(final Message message)
  {
    checkOwner(message);
    return message.isEmpty();
  }

  /**
   * @see MessageManagerType#getChunk(Message, String)
   */
  public AbstractChunk getChunk(final Message message, final String chunkName)
  {
    checkOwner(message);
    checkNull(chunkName);
    final AbstractChunk chunk = message.getChunk(chunkName);
    if (chunk == null)
    {
      return null;
    }
    return chunk;
  }

  /**
   * @see MessageManagerType#addChunk(Message, String, AbstractChunk)
   */
  public void addChunk(final Message message, final String chunkName,
      final AbstractChunk chunk)
  {
    checkOwner(message);
    checkOwner(chunk);
    checkNull(chunkName);
    if (message.getChunk(chunkName) != null)
    {
      throw new ChunkAlreadyExistException(chunkName);
    }
    message.addChunk(chunkName, chunk);
  }

  /**
   * @see MessageManagerType#addOrReplaceChunk(Message, String, AbstractChunk)
   */
  public AbstractChunk addOrReplaceChunk(final Message message,
      final String chunkName, final AbstractChunk chunk)
  {
    checkOwner(message);
    checkOwner(chunk);
    checkNull(chunkName);
    final AbstractChunk previous = message.getChunk(chunkName);
    message.addChunk(chunkName, chunk);
    return previous;
  }

  /**
   * @see MessageManagerType#removeChunk(Message, String)
   */
  public AbstractChunk removeChunk(final Message message, final String chunkName)
  {
    checkOwner(message);
    checkNull(chunkName);
    final AbstractChunk chunk = message.getChunk(chunkName);
    if (chunk == null)
    {
      throw new UnknownChunkException(chunkName);
    }
    message.removeChunk(chunkName);
    return chunk;
  }

  /**
   * @see MessageManagerType#getChunkNameIterator(Message)
   */
  public Iterator<String> getChunkNameIterator(final Message message)
  {
    checkOwner(message);
    checkOwner(message);
    final Iterator<String> iter = message.getNameIterator();
    if (!iter.hasNext())
    {
      return EmptyIterator.EMPTY_STRING_ITERATOR;
    }
    return new IteratorNoRemove<String>(iter);
  }

  // Sub message manipulation --------------------------------------------------

  /**
   * @see MessageManagerType#addSubMessage(Message, Message)
   */
  public void addSubMessage(final Message message, final Message subMessage)
  {
    checkOwner(message);
    checkOwner(subMessage);
    checkSubMessage(message, subMessage);
    message.addSubMessage(subMessage);
  }

  /**
   * @see MessageManagerType#getSubMessageIterator(Message)
   */
  public Iterator<Message> getSubMessageIterator(final Message message)
  {
    checkOwner(message);
    final Iterator<Message> subMessageIterator = message
        .getSubMessageIterator();
    if (!subMessageIterator.hasNext())
    {
      return EMPTY_SUB_MESSAGE_ITERATOR;
    }
    return new IteratorNoRemove<Message>(subMessageIterator);
  }

  /**
   * @see MessageManagerType#removeSubMessage(Message, Message)
   */
  public void removeSubMessage(final Message message, final Message subMessage)
  {
    checkOwner(message);
    checkOwner(subMessage);
    final boolean removed = message.removeSubMessage(subMessage);
    if (!removed)
    {
      throw new IllegalArgumentException(
          "Given subMessage is not a sub message of the given message");
    }
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  /**
   * Checks the owner of a message
   */
  protected final void checkOwner(final Message message)
  {
    checkNull(message);
    if (message.getOwner() != this)
    {
      throw new InvalidReference(
          "Message is not owned by this message manager", null);
    }
  }

  /**
   * Checks the owner of a chunk
   */
  protected final void checkOwner(final AbstractChunk<?> chunk)
  {
    checkNull(chunk);
    if (chunk.getOwner() != this)
    {
      throw new InvalidReference("Chunk is not owned by this message manager",
          null);
    }
  }

  /**
   * Check if the given chunk name is null
   * 
   * @param chunkName
   */
  protected final void checkNull(final String chunkName)
  {
    if (chunkName == null)
    {
      throw new NullPointerException("Given chunk name is null");
    }
  }

  /**
   * Check if the given chunk is null
   * 
   * @param chunk
   */
  protected final void checkNull(final AbstractChunk chunk)
  {
    if (chunk == null)
    {
      throw new NullPointerException("Given chunk is null");
    }
  }

  /**
   * Checks if the given message is null
   * 
   * @param message
   */
  protected final void checkNull(final Message message)
  {
    if (message == null)
    {
      throw new NullPointerException("Given message reference is null");
    }
  }

  /**
   * Checks if there are any problems in message inclusion
   * 
   * @param message
   * @param subMessage
   */
  protected final void checkSubMessage(final Message message,
      final Message subMessage)
  {
    if (!check)
      return;
    if (message == subMessage)
    {
      throw new IllegalArgumentException("message and subMessage are equals");
    }
    if (subMessage.isSubMessage(message))
    {
      throw new IllegalArgumentException(
          "message is a sub message of subMessage");
    }
    final Iterator<Message> iterator = subMessage.getSubMessageIterator();
    while (iterator.hasNext())
    {
      checkSubMessage(message, iterator.next());
    }
  }

  /**
   * create a chunk factory handling message of type <code>T</code>
   * 
   * @param <T> Type of the chunk
   * @param chunkType
   * @return the created chunk factory
   */
  protected <T extends AbstractChunk<T>> ChunkFactoryReference<T> createChunkFactory(
      final Class<T> chunkType)
  {
    try
    {
      return new BasicChunkFactoryImpl<T>(chunkType, this);
    }
    catch (Exception e)
    {
      throw new IllegalArgumentException(e);
    }
  }

  /**
   * @param message
   */
  protected void recycleMessageInstance(final Message message)
  {
  }
}
