/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.message;

import java.util.Iterator;

import org.objectweb.fractal.fraclet.annotation.annotations.Interface;

/**
 * Interface to creates, deletes and modify dream messages.
 */
@Interface(name = MessageManagerType.ITF_NAME)
public interface MessageManagerType
{
  /** The commonly used name of refer to this interface. */
  String ITF_NAME = "message-manager";

  // ---------------------------------------------------------------------------
  // Message creation methods
  // ---------------------------------------------------------------------------

  /**
   * Creates a new message and returns a reference to the newly created message.
   * 
   * @return a reference to a newly created message.
   */
  Message createMessage();

  /**
   * Delete the message referenced by the given reference. If the message has
   * been duplicated by reference (see {@link #duplicateMessage}), this
   * operation releases the given reference. The message is effectively deleted
   * only if all its references have been released. In this case, chunks in the
   * message are removed and recycled, so as message instance.
   * 
   * @param message the reference of the message to delete.
   * @throws InvalidReference if the given message reference is invalid.
   */
  void deleteMessage(Message message);

  /**
   * Duplicates a message. There are two kinds of duplication:
   * <ul>
   * <li>by <b>reference </b>. The returned Message is <i>the same object </i>
   * than the given one. The manager keeps track of the number of message
   * references, so that the {@link #deleteMessage} operation only deletes the
   * message when all the duplicated references have been released </li>
   * <li>by <b>value </b> (cloning). The returned message is a new message
   * which contains a copy of each chunk and each enclosed message of the
   * original message</li>
   * </ul>
   * 
   * @param message the reference of the message to be duplicated.
   * @param clone the duplication mode. <code>false</code> for reference
   *          duplication, <code>true</code> for value duplication.
   * @return the reference of the duplicated message.
   * @throws InvalidReference if the given message reference is invalid.
   */
  Message duplicateMessage(Message message, boolean clone);

  // ---------------------------------------------------------------------------
  // Chunk creation methods
  // ---------------------------------------------------------------------------

  /**
   * Returns a reference to a chunk factory for the given chunk type.
   * 
   * @param <T> the chunk type.
   * @param chunkType the class object of the required chunk type.
   * @return a reference to a chunk factory for the given chunk type.
   */
  <T extends AbstractChunk<T>> ChunkFactoryReference<T> getChunkFactory(
      Class<T> chunkType);

  /**
   * Creates a chunk using the chunk factory referenced by the given reference.
   * 
   * @param <T> the chunk type.
   * @param chunkFactory a chunk factory reference.
   * @return a chunk instance of the given type.
   * @throws InvalidReference if the given chunk factory reference is invalid.
   */
  <T extends AbstractChunk<T>> T createChunk(
      ChunkFactoryReference<T> chunkFactory);

  /**
   * Deletes the given chunk.
   * 
   * @param chunk the chunk to delete.
   * @throws InvalidReference if the given chunk is invalid.
   */
  void deleteChunk(AbstractChunk<?> chunk);

  /**
   * Clones the given chunk.
   * 
   * @param <T> the type of the chunk to clone.
   * @param chunk the chunk to clone.
   * @return the cloned chunk.
   * @throws InvalidReference if the given chunk is invalid.
   */
  <T extends AbstractChunk<T>> T cloneChunk(T chunk);

  // ---------------------------------------------------------------------------
  // Message manipulation methods
  // ---------------------------------------------------------------------------

  // Chunk manipulation --------------------------------------------------------

  /**
   * Returns <code>true</code> if the message is empty (ie. contains no chunk
   * and no sub message).
   * 
   * @param message the reference of a message
   * @return <code>true</code> if the message is empty.
   * @throws InvalidReference if the given message reference is invalid.
   */
  boolean isEmpty(Message message);

  /**
   * Returns the chunk in the given message with the given name.
   * 
   * @param message the reference of the message
   * @param chunkName the name of the chunk.
   * @return a chunk of the type specified by the chunk name, or
   *         <code>null</code> if no chunk can be found.
   * @throws InvalidReference if the given message reference is invalid.
   */
  AbstractChunk getChunk(Message message, String chunkName);

  /**
   * Adds a chunk in the given message. If chunk with the given name already
   * exist in the message, a {@link ChunkAlreadyExistException} is thrown.
   * 
   * @param message the reference of the message in which the chunk will be
   *          added.
   * @param chunkName the name of the chunk.
   * @param chunk the chunk to add
   * @throws InvalidReference if the given message reference is invalid.
   * @throws ChunkAlreadyExistException if a chunk with the same name already
   *           exist in the given message.
   */
  void addChunk(Message message, String chunkName, AbstractChunk chunk);

  /**
   * Adds a chunk in the given message. If a chunk with the given name already
   * exist in the message, it is replaced and returned.
   * 
   * @param message the reference of the message in which the message will be
   *          added.
   * @param chunkName the name of the chunk.
   * @param chunk the chunk to add
   * @return the previous chunk with the given name, or <code>null</code>.
   * @throws InvalidReference if the given message reference is invalid.
   */
  AbstractChunk addOrReplaceChunk(Message message, String chunkName,
      AbstractChunk chunk);

  /**
   * Removes a chunk from the given message. If no chunk with the given name can
   * be found in the message, a {@link UnknownChunkException} is thrown.
   * 
   * @param message the reference of the message from which the chunk will be
   *          removed.
   * @param chunkName the name of the chunk.
   * @return the removed chunk.
   * @throws InvalidReference if the given message reference is invalid.
   * @throws UnknownChunkException if no chunk with the given name can be found
   *           in the message.
   */
  AbstractChunk removeChunk(Message message, String chunkName);

  /**
   * Returns an iterator over the chunk names of the given message.
   * 
   * @param message a message reference.
   * @return an iterator over the chunk names of the given message.
   * @throws InvalidReference if the given message reference is invalid.
   */
  Iterator<String> getChunkNameIterator(Message message);

  // Sub message manipulation --------------------------------------------------

  /**
   * Adds a sub message to a message.
   * 
   * @param message the message where the sub message will be added.
   * @param subMessage the sub message.
   * @throws InvalidReference if the given message reference or sub message
   *           reference are invalid.
   * @throws IllegalArgumentException if this operation may creates a cycle (if
   *           <code>message</code> is already a sub message of
   *           <code>subMessage</code>). Since this check is expensive, some
   *           implementations may not perform it.
   */
  void addSubMessage(Message message, Message subMessage);

  /**
   * Removes a sub message from a message.
   * 
   * @param message the message from which the sub message will be removed.
   * @param subMessage the sub message.
   * @throws InvalidReference if the given message reference or sub message
   *           reference are invalid.
   * @throws IllegalArgumentException if <code>subMessage</code> is not a sub
   *           message of <code>message</code>.
   */
  void removeSubMessage(Message message, Message subMessage);

  /**
   * Returns an iterator over sub messages of the given message.
   * 
   * @param message a message reference.
   * @return an iterator over the sub messages of the given message.
   * @throws InvalidReference if the given message reference or sub message
   *           reference are invalid.
   */
  Iterator<Message> getSubMessageIterator(Message message);
}
