/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Pierre Garcia
 * Contributor(s): Valerio Schiavoni 
 */

package org.objectweb.dream.message.codec.compression;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.ref.Reference;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import org.objectweb.dream.message.codec.CodecInputOutput;
import org.objectweb.dream.message.codec.MessageCodec;
import org.objectweb.dream.pool.Recyclable;

/**
 * Zip Codec used in the {@link MessageCodecHelperZipStream} to zip/unzip
 * message. This codec uses another to delegate the real encoding/decoding.
 * 
 * In consequence, the method <code>setDelegatedCodec</code> must be called before
 * {@link MessageCodec#encode(CodecInputOutput, org.objectweb.dream.message.Message)}
 * 
 * @author pgarcia
 */
public class ZipCodecInputOutput implements Recyclable, CodecInputOutput
{
  private final static boolean              DEBUG = false;

  /**
   * InputStream linked to the codec
   */
  protected Object                          input;

  /**
   * OutputStream linked to the codec
   */
  protected Object                          output;

  /**
   * Stream used to unzip
   */
  protected ZipInputStream                  zipInputStream;

  /**
   * Stream used to zip
   */
  protected ZipOutputStream                 zipOutputStream;

  /**
   * Codec to which the lower level decoding/encoding is delegated
   */
  protected CodecInputOutput                delegatedCodec;

  /**
   * Reference used by the pool of objets
   */
  protected Reference<? extends Recyclable> recyclableReference;

  /**
   * Default constructor
   * @param delegatedCodec
   * @throws IOException
   */
  public ZipCodecInputOutput(CodecInputOutput delegatedCodec) throws IOException
  {
    this.delegatedCodec = delegatedCodec;
    zipOutputStream = new ZipOutputStream((OutputStream) delegatedCodec
        .getOutput());
    output = zipOutputStream;
    if (DEBUG)
    {
      System.out.println("zipOutputStream created");
    }
  }
  
  // ---------------------------------------------------------------------------
  // Implementation of the CodecInputOutput interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.message.codec.CodecInputOutput#close()
   */
  public void close() throws IOException
  {
    if (zipOutputStream != null)
    {
      // the next line is needed by ZipOutputStream
      zipOutputStream.putNextEntry(new ZipEntry("zipOutputStream closing"));
      zipOutputStream.close();
    }
    if (zipInputStream != null)
    {
      zipInputStream.close();
    }
  }

  /**
   * @see org.objectweb.dream.message.codec.CodecInputOutput#getInput()
   */
  public Object getInput() throws IOException
  {
    return input;
  }

  /**
   * @see org.objectweb.dream.message.codec.CodecInputOutput#getOutput()
   */
  public Object getOutput() throws IOException
  {
    return output;
  }

  /**
   * @see org.objectweb.dream.message.codec.CodecInputOutput#setInput(java.lang.Object)
   */
  public void setInput(Object input)
  {
    this.input = input;
  }

  /**
   * @see org.objectweb.dream.message.codec.CodecInputOutput#setOutput(java.lang.Object)
   */
  public void setOutput(Object output)
  {
    if (DEBUG)
    {
      System.out.println("setOutput : " + output.getClass().getSimpleName());
    }
    this.output = output;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Recyclable interface
  // ---------------------------------------------------------------------------

  public void recycle()
  {
    this.delegatedCodec = null;
    this.zipInputStream = null;
    this.zipOutputStream = null;
  }

  /**
   * @see Recyclable#setReference(Reference)
   */
  public void setReference(Reference<? extends Recyclable> reference)
  {
    this.recyclableReference = reference;
  }

  /**
   * @see Recyclable#getReference()
   */
  public Reference<? extends Recyclable> getReference()
  {
    return recyclableReference;
  }

  // ---------------------------------------------------------------------------
  // Codec Specific methods
  // ---------------------------------------------------------------------------

  public CodecInputOutput getDelegatedCodec()
  {
    return delegatedCodec;
  }

  public ZipInputStream getZipInputStream() throws IOException
  {
    if (zipInputStream == null)
    {
      if (delegatedCodec == null)
        throw new InternalError(
            "the zip codec needs another codec to delegate the \"real encoding\"");
      zipInputStream = new ZipInputStream((InputStream) delegatedCodec
          .getInput());
      input = zipInputStream;
      if (DEBUG)
      {
        System.out.println("zipInputStream created");
      }
    }
    return zipInputStream;
  }

  public ZipOutputStream getZipOutputStream()
  {
    return zipOutputStream;
  }
}
