/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s):
 */

package org.objectweb.dream.util;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.dream.message.Message;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.util.Fractal;
import org.objectweb.util.monolog.api.Logger;

/**
 * This class defines utility methods.
 */
public final class Util
{
  /**
   * Private constructor
   */
  private Util()
  {
  }

  /** A null object. Useful for Map values when only the key is meaningful. */
  public static final Object NULL_OBJECT = new Object();

  /**
   * Retrieves the sub component with the specified name from the content of a
   * composite. Name must have the form "comp1/comp2/comp3".
   * 
   * @param cc the content controller of the composite.
   * @param name the name of the component to be retrieved.
   * @return the subcomponent with the specified name, <code>null</code> if no
   *         component with the specified name exists.
   */
  public static Component getComponentByName(ContentController cc, String name)
  {
    String[] compNames = name.split("/");
    // The component to be returned
    Component c = null;
    for (int i = 0; i < compNames.length; i++)
    {
      c = null;
      Component[] comps = cc.getFcSubComponents();
      for (int j = 0; j < comps.length; j++)
      {
        Component component = comps[j];
        try
        {
          String tempName = Fractal.getNameController(component).getFcName();
          if (tempName.equals(compNames[i]))
          {
            c = component;
            cc = Fractal.getContentController(component);
            break;
          }
        }
        catch (NoSuchInterfaceException e)
        {
          // Do nothing
        }
      }
      if (c == null)
      {
        return null;
      }
    }
    return c;
  }

  /**
   * This method implements a "partial duplication" of messages. It creates a
   * new message and adds into this message chunks and sub messages of the
   * duplicated message. Sub messages a duplicated by reference (see
   * {@link MessageManagerType#duplicateMessage(Message, boolean)})
   * 
   * @param message the message to be "partially duplicated".
   * @param messageManagerItf the message manager to be used.
   * @param logger the logger to be used to log errors.
   * @return a partial duplicate of the message.
   */
  public static Message partiallyDuplicateMessage(Message message,
      MessageManagerType messageManagerItf, Logger logger)
  {
    Message duplicatedMessage = messageManagerItf.createMessage();
    Iterator<String> iter = messageManagerItf.getChunkNameIterator(message);
    while (iter.hasNext())
    {
      String chunkName = iter.next();
      messageManagerItf.addChunk(duplicatedMessage, chunkName,
          messageManagerItf.getChunk(duplicatedMessage, chunkName));
    }
    Iterator<Message> subMsgIterator = messageManagerItf
        .getSubMessageIterator(message);
    while (subMsgIterator.hasNext())
    {
      Message subMessage = subMsgIterator.next();
      messageManagerItf.addSubMessage(duplicatedMessage, messageManagerItf
          .duplicateMessage(subMessage, false));
    }
    return duplicatedMessage;
  }

  /**
   * Writes an array of bytes on a <code>DataOutput</code> object.
   * <p>
   * <i>Note: </i> if the array is <code>null</code>,<code>-1</code> is
   * written.
   * 
   * @param out the output on which to write.
   * @param array the array to write.
   * @throws IOException if an error occurs.
   */
  public static void writeExternalByteArray(DataOutput out, byte[] array)
      throws IOException
  {
    if (array == null)
    {
      out.writeInt(-1);
    }
    else
    {
      int length = array.length;
      out.writeInt(length);
      out.write(array);
    }
  }

  /**
   * Reads and returns an array of bytes from an <code>ObjectInput</code>
   * object.
   * 
   * @param in the input from which to read.
   * @return an array of byte.
   * @throws IOException if an error occurs.
   */
  public static byte[] readExternalByteArray(DataInput in) throws IOException
  {
    int length = in.readInt();
    if (length == -1)
    {
      return null;
    }
    byte[] toReturn = new byte[length];
    in.readFully(toReturn);
    return toReturn;
  }

  /**
   * Writes an array of integer on a <code>DataOutput</code> object.
   * <p>
   * <i>Note: </i> if the array is <code>null</code>,<code>-1</code> is
   * written.
   * 
   * @param out the output on which to write.
   * @param array the array to write.
   * @throws IOException if an error occurs.
   */
  public static void writeExternalIntArray(DataOutput out, int[] array)
      throws IOException
  {
    if (array == null)
    {
      out.writeInt(-1);
    }
    else
    {
      int length = array.length;
      out.writeInt(length);
      for (int i = 0; i < length; i++)
      {
        out.writeInt(array[i]);
      }
    }
  }

  /**
   * Reads and returns an array of integer from a <code>DataInput</code>
   * object.
   * 
   * @param in the input from which to read.
   * @return an array of integer.
   * @throws IOException if an error occurs.
   */
  public static int[] readExternalIntArray(DataInput in) throws IOException
  {
    int length = in.readInt();
    if (length == -1)
    {
      return null;
    }
    int[] toReturn = new int[length];
    for (int i = 0; i < length; i++)
    {
      toReturn[i] = in.readInt();
    }
    return toReturn;
  }

  /**
   * Writes an array of long on a <code>DataOutput</code> object.
   * <p>
   * <i>Note: </i> if the array is <code>null</code>,<code>-1</code> is
   * written.
   * 
   * @param out the output on which to write.
   * @param array the array to write.
   * @throws IOException if an error occurs.
   */
  public static void writeExternalLongArray(DataOutput out, long[] array)
      throws IOException
  {
    if (array == null)
    {
      out.writeInt(-1);
    }
    else
    {
      int length = array.length;
      out.writeInt(length);
      for (int i = 0; i < length; i++)
      {
        out.writeLong(array[i]);
      }
    }
  }

  /**
   * Reads and returns an array of long from a <code>DataInput</code> object.
   * 
   * @param in the input from which to read.
   * @return an array of long.
   * @throws IOException if an error occurs.
   */
  public static long[] readExternalLongArray(DataInput in) throws IOException
  {
    int length = in.readInt();
    if (length == -1)
    {
      return null;
    }
    long[] toReturn = new long[length];
    for (int i = 0; i < length; i++)
    {
      toReturn[i] = in.readLong();
    }
    return toReturn;
  }

  /**
   * Writes an array of objects on an <code>ObjectOutput</code> object.
   * <p>
   * <i>Note: </i> if the array is <code>null</code>,<code>-1</code> is
   * written.
   * 
   * @param out the output on which to write.
   * @param array the array to write.
   * @throws IOException if an error occurs.
   */
  public static void writeExternalObjectArray(ObjectOutput out, Object[] array)
      throws IOException
  {
    if (array == null)
    {
      out.writeInt(-1);
    }
    else
    {
      int length = array.length;
      out.writeInt(length);
      for (int i = 0; i < length; i++)
      {
        writeObject(out, array[i]);
      }
    }
  }

  /**
   * Reads and returns an array of objects from an <code>ObjectInput</code>
   * object.
   * 
   * @param in the input from which to read.
   * @return an array of objects.
   * @throws IOException if an error occurs.
   * @throws ClassNotFoundException if the class of read objects is not found.
   */
  public static Object[] readExternalObjectArray(ObjectInput in)
      throws IOException, ClassNotFoundException
  {
    int length = in.readInt();
    if (length == -1)
    {
      return null;
    }
    Object[] toReturn = new Object[length];
    for (int i = 0; i < length; i++)
    {
      toReturn[i] = readObject(in);
    }
    return toReturn;
  }

  public static <T> Collection<T> readExternalObjectArray(ObjectInput in,
      Class<T> c) throws IOException, ClassNotFoundException
  {

    int length = in.readInt();
    if (length == -1)
    {
      return null;
    }
    List<T> toReturn = new ArrayList<T>(length);
    for (int i = 0; i < length; i++)
    {
      toReturn.add(i, (T) readObject(in));
    }
    return toReturn;
  }

  /**
   * Writes an array of externalizable objects on an <code>ObjectOutput</code>
   * object.
   * <p>
   * <i>Note: </i> if the array is <code>null</code>,<code>-1</code> is
   * written.
   * 
   * @param out the output on which to write.
   * @param array the array of externalizable objects to write.
   * @throws IOException if an error occurs.
   */
  public static void writeExternalExternalizableObjectArray(ObjectOutput out,
      Object[] array) throws IOException
  {
    if (array == null)
    {
      out.writeInt(-1);
    }
    else
    {
      int length = array.length;
      out.writeInt(length);
      for (int i = 0; i < length; i++)
      {
        if (array[i] == null)
        {
          // write isNull == true
          out.writeBoolean(true);
        }
        else
        {
          out.writeBoolean(false);
          ((Externalizable) array[i]).writeExternal(out);
        }
      }
    }
  }

  /**
   * Writes an array of strings on an <code>ObjectOutput</code> object.
   * <p>
   * <i>Note: </i> if the array is <code>null</code>,<code>-1</code> is
   * written.
   * 
   * @param out the output on which to write.
   * @param array the array to write.
   * @throws IOException if an error occurs.
   */
  public static void writeExternalStringArray(ObjectOutput out, String[] array)
      throws IOException
  {
    if (array == null)
    {
      out.writeInt(-1);
    }
    else
    {
      int length = array.length;
      out.writeInt(length);
      for (int i = 0; i < length; i++)
      {
        if (array[i] == null)
        {
          // write isNull == true
          out.writeBoolean(true);
        }
        else
        {
          out.writeBoolean(false);
          out.writeUTF(array[i]);
        }
      }
    }
  }

  /**
   * Reads and returns an array of strings from an <code>ObjectInput</code>
   * object.
   * 
   * @param in the input from which to read.
   * @return an array of strings.
   * @throws IOException if an error occurs.
   * @throws ClassNotFoundException if the class of read objects is not found.
   */
  public static String[] readExternalStringArray(ObjectInput in)
      throws IOException, ClassNotFoundException
  {
    int length = in.readInt();
    if (length == -1)
    {
      return null;
    }
    String[] toReturn = new String[length];
    for (int i = 0; i < length; i++)
    {
      if (!in.readBoolean())
      {
        // isNull == false
        toReturn[i] = in.readUTF();
      }
    }
    return toReturn;
  }

  /**
   * Writes a two-dimensions array of objects on an <code>ObjectOutput</code>
   * object.
   * <p>
   * <i>Note: </i> if the array (of one of its sub-arrays) is <code>null</code>,
   * <code>-1</code> is written.
   * 
   * @param out the output on which to write.
   * @param twoDimensionsArray the two-dimensions array to write.
   * @throws IOException if an error occurs.
   */
  public static void writeExternalTwoDimensionsObjectArray(ObjectOutput out,
      Object[][] twoDimensionsArray) throws IOException
  {
    if (twoDimensionsArray == null)
    {
      out.writeInt(-1);
    }
    else
    {
      int length = twoDimensionsArray.length;
      out.writeInt(length);
      for (int i = 0; i < length; i++)
      {
        writeExternalObjectArray(out, twoDimensionsArray[i]);
      }
    }
  }

  /**
   * Reads and returns a two-dimensions array of objects from an
   * <code>ObjectInput</code> object.
   * 
   * @param in the input from which to read.
   * @return a two-dimensions array of objects.
   * @throws IOException if an error occurs.
   * @throws ClassNotFoundException if the class of read objects is not found.
   */
  public static Object[][] readExternalTwoDimensionsObjectArray(ObjectInput in)
      throws IOException, ClassNotFoundException
  {
    int length = in.readInt();
    if (length == -1)
    {
      return null;
    }
    Object[][] toReturn = new Object[length][];
    for (int i = 0; i < length; i++)
    {
      toReturn[i] = readExternalObjectArray(in);
    }
    return toReturn;
  }

  /**
   * Writes a two-dimensions array of strings on an <code>ObjectOutput</code>
   * object.
   * <p>
   * <i>Note: </i> if the array (of one of its sub-arrays) is <code>null</code>,
   * <code>-1</code> is written.
   * 
   * @param out the output on which to write.
   * @param twoDimensionsArray the two-dimensions array to write.
   * @throws IOException if an error occurs.
   */
  public static void writeExternalTwoDimensionsStringArray(ObjectOutput out,
      String[][] twoDimensionsArray) throws IOException
  {
    if (twoDimensionsArray == null)
    {
      out.writeInt(-1);
    }
    else
    {
      int length = twoDimensionsArray.length;
      out.writeInt(length);
      for (int i = 0; i < length; i++)
      {
        writeExternalStringArray(out, twoDimensionsArray[i]);
      }
    }
  }

  /**
   * Reads and returns a two-dimensions array of strings from an
   * <code>ObjectInput</code> object.
   * 
   * @param in the input from which to read.
   * @return a two-dimensions array of strings.
   * @throws IOException if an error occurs.
   * @throws ClassNotFoundException if the class of read objects is not found.
   */
  public static String[][] readExternalTwoDimensionsStringArray(ObjectInput in)
      throws IOException, ClassNotFoundException
  {
    int length = in.readInt();
    if (length == -1)
    {
      return null;
    }
    String[][] toReturn = new String[length][];
    for (int i = 0; i < length; i++)
    {
      toReturn[i] = readExternalStringArray(in);
    }
    return toReturn;
  }

  /**
   * Write the given object on the given {@link ObjectOutput}. If the given
   * output is an {@link ObjectOutputStream}, then use the
   * {@link ObjectOutputStream#writeUnshared(Object) }method, else use the
   * normal {@link ObjectOutput#writeObject(Object) }method.
   * 
   * @see ObjectOutput#writeObject(Object)
   */
  public static void writeObject(ObjectOutput output, Object o)
      throws IOException
  {
    if (output instanceof ObjectOutputStream)
    {
      // the writeUnshared operation is not supported by Xstream
      // ObjectOutputStream
      try
      {
        ((ObjectOutputStream) output).writeUnshared(o);
      }
      catch (UnsupportedOperationException e)
      {
        output.writeObject(o);
      }
    }
    else
    {
      output.writeObject(o);
    }
  }

  /**
   * Read the given object from the given {@link ObjectInput}. If the given
   * input is an {@link ObjectInputStream}, then use the
   * {@link ObjectInputStream#readUnshared() }method, else use the normal
   * {@link ObjectInput#readObject() }method.
   * 
   * @see ObjectInput#readObject()
   */
  public static Object readObject(ObjectInput input) throws IOException,
      ClassNotFoundException
  {
    if (input instanceof ObjectInputStream)
    {
      // the readUnshared operation is not supported by Xstream
      // ObjectInputStream
      try
      {
        return ((ObjectInputStream) input).readUnshared();
      }
      catch (UnsupportedOperationException e)
      {
        return input.readObject();
      }
    }
    return input.readObject();
  }
}
